--global tabel
MoreStopwatches = {};
MoreStopwatches.timerList = {};

--debug stuff
MoreStopwatches.debugEnabled = false;
local function debug(...)
	if MoreStopwatches.debugEnabled then
		print("MoreStopwachtes Debug:", ...);
	end;
end;

--do our stuff once saved vars are loaded
local startup = CreateFrame("Frame");
startup:RegisterEvent("ADDON_LOADED");
startup:SetScript("OnEvent",function(self, event, addonName)
	if addonName == "MoreStopwatches" then
		--locals
		local FRAME_PREFIX = "MoreStopwatches_";
		local timerID = 0;

		--init saved vars on first login
		if not MoreStopwatchesSave then
			MoreStopwatchesSave = {};
		end;

		--for UTF8sub function below
		local function chsize(char)
			if not char then
				return 0;
			elseif char>240 then
				return 4;
			elseif char>225 then
				return 3;
			elseif char>192 then
				return 2;
			else
				return 1;
			end;
		end;

		--UTF8 aware strsub function
		--taken from http://wowprogramming.com/snippets/UTF-8_aware_stringsub_7
		local function utf8sub(str, startChar, numChars)
			local startIndex = 1;
			while startChar>1 do
				local char = strbyte(str,startIndex);
				startIndex = startIndex + chsize(char);
				startChar = startChar - 1;
			end;

			local currentIndex = startIndex;
			while ( numChars > 0 ) and ( currentIndex <= #str ) do
				local char = strbyte(str, currentIndex);
				currentIndex = currentIndex + chsize(char);
				numChars = numChars -1;
			end;

			return str:sub(startIndex, currentIndex - 1);
		end;

		--prepends 0 for low ints: 9 -> 09
		local function IntToStr(i)
			if ( i < 10 ) then
				return "0" .. tostring(i);
			else
				return tostring(i);
			end;
		end;

		--Create normal timer: /timer Test Timer Name
		--Create timer, which will be saved between sessions: /timer true Test Timer Name
		--Create timer, with starting value of 3 minutes: /timer 3 Test Timer Name
		--Create timer, with starting value of 3 minutes, which will be saved between sessions: /timer true 3 Test Timer Name OR /timer 3 true Test Timer Name
		local paramsPrototype = {
			__newindex = function(tab, key, value)
				if ( tab.timerName == "" ) then
					if ( value == "true" ) then
						rawset(tab, "saveBetweenSessions", true);
						debug("Timer will be saved between sessions.");
					elseif ( value == "false") then
						rawset(tab, "saveBetweenSessions", false);
						debug("Timer won't be saved between sessions.");
					elseif ( tonumber(value) ) then
						rawset(tab, "setTimer", tonumber(value));
						debug("Timer set to:", value);
					elseif ( type(value) == "string" ) then
						--first letter of name upper case
						local timerName = strupper(utf8sub(value, 1, 1)) .. utf8sub(value, 2, strlenutf8(value));
						rawset(tab, "timerName", timerName);
						
						--default for saveBetweenSessions is true
						if ( rawget(tab, "saveBetweenSessions") == nil ) then
							rawset(tab, "saveBetweenSessions", true);
						end;
						
						debug("Timername:",timerName,"Value was:",value);
					end;
				elseif ( type(value) == "string" ) then
					local timerName = format("%s %s", tab.timerName, value);
					rawset(tab, "timerName", timerName);
					debug("New timer name:",timerName,"Value was:",value);
				end;
			end,
		};

		--patterns to find strings/numbers at the begging of a slash command
		local pattern = {
			number = "^(%d+%.?%d*)%s+(.*)$",--get x: first number (int or float) and y: rest after white-space after that number
			string = "^(%S*)%s*(.-)$",--get x: first (whole) string and y: rest after white-space after that string
		};

		--returns a table of parameters, containing timerName=title of the timer, [setTimer=the time - in minutes - the timer should be set to, default:0], [saveBetweenSessions=true if timer state should be saved between sessions, default: true]
		local function GetParameters(msg)
			local params = setmetatable( { timerName = "", }, paramsPrototype );

			-- #parameters - 1
			local _, spaceCount = gsub(msg, "%s+", 1);

			local command, commandRest = nil, msg;

			--iterate over paramters and add them to the return (meta)table
			for i = 1, spaceCount + 1 do
				if ( (i == 1) or (i == 2) ) and ( not params.setTimer ) and ( commandRest:match(pattern.number) ) then
					command, commandRest = commandRest:match(pattern.number);
					debug("Paramter Number",i,command,"---",commandRest);
				else
					command, commandRest = commandRest:match(pattern.string);
					debug("Paramter String",i,command,"---",commandRest);
				end;

				if command then
					params[i] = command;
				end;
			end;

			return setmetatable(params, nil);
		end;

		--slash command function
		function MoreStopwatches.Slash(msg, status)
			debug("Slash msg:",msg," status:",status)
			
			--hide blizzard stopwatch
			if StopwatchFrame then
				StopwatchFrame:Hide();
			end;

			--paramters table contains: timerName[, saveBetweenSessions, setTimer]
			local params = GetParameters(msg);

			--initialize all time manger functions (which are adapted from Blizzard_TimeManager.lua) on first /timer call
			MoreStopwatches.Init();

			--use the command parameter as name for the timer, or generate a default name when called with '/timer'
			local timerName;
			if ( params.timerName == "" ) then
				timerID = timerID + 1;
				timerName = "Stopwatch" .. IntToStr(timerID);
			else
				timerName = params.timerName;
			end;

			local timer = MoreStopwatches.timerList[timerName];
			
			if timer then
				--show the timer if the timerName is already in use...
				timer:Show();
				debug("Timer already exists. Reusing.");
			else
				--...or create a new timer if not
				 timer = CreateFrame("Frame", FRAME_PREFIX .. timerName, UIParent, "MoreStopwatchesTemplate");
				 MoreStopwatches.timerList[timerName] = timer;
				 debug("New timer created.");
			end;

			--set a specific time (if user requested that) [in minutes]
			if params.setTimer then
				_G[timer:GetName() .. "StopwatchTicker"].timer = ( params.setTimer * 60 );
				MoreStopwatches_StopwatchTemplateTicker_Update(timer);
				debug("Timer set to requested time:", params.setTimer*60);
			end;

			if params.saveBetweenSessions then
				--register this stopwatch to be saved between sessions
				timer.saveBetweenSessions = true;
				_G[timer:GetName() .. "StopwatchTabFrameStopwatchTitle"]:SetText(timerName);
			else
				timer.saveBetweenSessions = nil;
				
				--show that it is not registered with a star in the label
				_G[timer:GetName() .. "StopwatchTabFrameStopwatchTitle"]:SetText(timerName .. "*");
			end;

			--only start the timer if it was not paused on save, or if we are called by user (then status is a table)
			if ( type(status)=="table" ) or ( status == "PLAYING" ) then
				MoreStopwatches_StopwatchTemplate_Play(_G[timer:GetName() .. "StopwatchPlayPauseButton"]);
				debug("Timer started on create:",timerName);
			end;
		end;

		--hook the blizzard /timer (also /sw and /stopwatch) slash command, this is where we do our magic
		hooksecurefunc(SlashCmdList, "STOPWATCH", MoreStopwatches.Slash);

		if not (MoreStopwatchesSave.savedTimers and MoreStopwatchesSave.time) then
			MoreStopwatchesSave.savedTimers = {};
			MoreStopwatchesSave.time = time();
		end;
		
		setmetatable(MoreStopwatchesSave.savedTimers, {
			__index = function(tab, key)
				rawset(tab, key, {});
				return rawget(tab, key);
			end
		});
		
		--restore saved stopwatches
		if MoreStopwatchesSave.savedTimers and MoreStopwatchesSave.time then
			for frameName, tab in pairs(MoreStopwatchesSave.savedTimers) do
				if ( type(tab.savedTime)=="number" ) then
					local restoreTime;
					
					if ( tab.playing ) then
						restoreTime= (time() - MoreStopwatchesSave.time) + tab.savedTime;
					else
						restoreTime = tab.savedTime;
					end;
					
					-- Slash command takes minutes
					restoreTime = restoreTime / 60;
					
					if ( restoreTime < 6000 ) then
						--strip frame-prefix to get timer label
						local label = gsub(frameName, FRAME_PREFIX, "");

						--simulate slash command (and pause button if it was not playing on save)
						MoreStopwatches.Slash(string.format("%f true %s", restoreTime, label), tab.playing and "PLAYING" or "PAUSED");
						
						--discard if restoreTime ~ 0
						local timer = MoreStopwatches.timerList[label];
						if ( ((restoreTime*60) < 0.2) and timer ) then
							MoreStopwatches_StopwatchTemplate_Clear(timer);
						end;
					else
						--saved time is too big, discard it
						MoreStopwatchesSave.savedTimers[frameName] = nil;
					end;
				else
					MoreStopwatchesSave.savedTimers[frameName] = nil;
				end;
			end;
		end;
		
		--everything set up - we are done here
		self:UnregisterAllEvents();
	end;
end);